#include <iostream>

#include "linearsolver.h"

using namespace Eigen;
using namespace std;

template<typename MatrixType, typename VectorType>
double check( const VectorType & sol, const MatrixType & lhs,
            const VectorType & rhs )
{
    return Util::abs( ( rhs - lhs * sol ).norm() / rhs.norm() );
}

template<typename T,
         typename LinearSolverType,
         typename MatrixType,
         typename VectorType,
         int Size>
double testLinearSolver( int repeat )
{
    double max_error = 0;

    cout << "size         " << Size << endl;
    cout << "repeat       " << repeat << endl;

    for( int count = 0; count < repeat; count++ )
    {
        MatrixType lhs(Size);
        VectorType sol(Size);
        T lhs_rows[Size][Size];
        for( int i = 0; i < Size * Size; i++)
            lhs_rows[i%Size][i/Size] = (rand()%2);
        int dependent_row = rand() % Size;
        for( int i = 0; i < Size; i++)
        {
            lhs_rows[dependent_row][i] = 0;
            for( int j = 0; j < Size; j++)
                lhs_rows[dependent_row][i] += static_cast<T>(1+j) * lhs_rows[j][i];
        }
        lhs.readRows( &lhs_rows[0][0] );
        VectorType some_vector(Size);
        for( int i = 0; i < Size; i++) some_vector(i) = (rand()%11) - 5;
        VectorType rhs = lhs * some_vector;
        LinearSolverType solver( lhs, rhs );
        if( solver.thereExistSolutions() )
        {
            max_error = Util::max( max_error,
                          check( solver.someSolution(), lhs, rhs ) );

            if( solver.dimSolutions() != 0 )
            {
                // let's construct a random parameter vector:
                T param[Size];
                for( int i = 0; i < solver.dimSolutions(); i++ )
                    param[i] = -10.0 + 20.0 * rand() / RAND_MAX;
                sol = solver.genericSolution( param );
                max_error = Util::max( max_error, check(sol, lhs, rhs ) );
            }
        }
    }
    cout << "max_error    " << max_error << endl;
    return max_error;
}

int main( int argc, char *argv[] )
{

	Vector4d v ( 2, 0, -1, 4 );
	double rows[4][4] = { { 2, 1, -1, 3}, {2, 7, 1, 5}, {-3, 2, 0, 3}, { 9, 13, 1, 10} };
	Matrix4d m;
	m.readRows( &rows[0][0] );
	cout << v << endl << m << endl << m * v << endl;
	cout << "det " << m.determinant() << endl;

#if 0
    Vector3d v1, v2;
    v1(0) = 1;
    v1(1) = 0;
    v1(2) = 1;

    v1.makeOrthoVector( &v2 );
    cout << v2;
#endif

#if 0
    cout << endl;
    cout << "************************************" << endl;
    cout << "* testing fixed-size linear solver *" << endl;
    cout << "************************************" << endl;

    testLinearSolver< double,
                      LinearSolver<double, 5>,
                      Matrix<double, 5>,
                      Vector<double, 5>,
                      5 > ( 100 );
#endif

#if 0
    cout << endl;
    cout << "**************************************" << endl;
    cout << "* testing dynamic-size linear solver *" << endl;
    cout << "**************************************" << endl;

    testLinearSolver< double,
                      LinearSolverX<double>,
                      MatrixX<double>,
                      VectorX<double>,
                      12 > (10);
#endif

#if 0
    cout << endl;
    cout << "**********************************************" << endl;
    cout << "* testing dynamic-size complex linear solver *" << endl;
    cout << "**********************************************" << endl;

    testLinearSolver< complex<double>,
                      LinearSolverX< complex<double> >,
                      MatrixX< complex<double> >,
                      VectorX< complex<double> >,
                      12 > (10);
#endif

#if 0
    double c[3][3] = { { 2.3, 4.5, 1.2 },
                       { -2.1, 1.5, 4.2 },
                       { 2.9, 2.3, 1.8 } };

/*    double c[3][3] = { { 2.3, 4.5, 6.8/7 },
                       { -2.1, 1.5, -0.6/7 },
                       { 2.9, 2.3, 5.2/7 } };*/

//    double c[3][3] = {{0,0,0},{0,0,0},{0,0,0}};

    Matrix3d mat(3);
    mat.readRows( &c[0][0] );
    cout << mat;
    Matrix3d inverse(3);
    if( mat.computeInverse( &inverse ) )
    {
        cout << "inverse:" <<endl;
        cout << inverse <<endl;
        cout << "check:" <<endl;
        cout << (inverse * mat) <<endl;
    }
    else
        cout << "non invertible" << endl;
#endif

#if 0
    Matrix2d matc;
    matc(0,0) = complex<double>( 1, 1);
    matc(1,0) = complex<double>( -2, 3);
    matc(0,1) = complex<double>( 3, -1);
    matc(1,1) = complex<double>( 4, -2);

    cout << matc <<endl;

    cout << matc.adjoint() <<endl;

    matc.replaceByAdjoint();

    cout << matc << endl;
#endif

#if 0
    double c3[3][3] = { { 2.3, 4.5, 1.2 },
                       { -2.1, 1.5, 4.2 },
                       { 2.9, 2.3, 1.8 } };

    Vector3d vec3 ( &c3[0][0] );
    Vector3d vec4 ( &c3[1][0] );

    Matrix3d mat3;
    mat3.readRows( &c3[0][0] );

    cout << vec3 << endl;
    cout << mat3 << endl;

    cout << mat3 * vec3 << endl;

    cout << cross( vec3, vec4 ) << endl;
#endif

    return 1;
}
